#!/usr/bin/env python


from lxml import etree
from typing import Any

from ruia.exceptions import IgnoreThisItem, InvalidFuncType
from ruia.field import BaseField
from ruia.request import Request


class ItemMeta(type):
    """
    Metaclass for an item
    """

    def __new__(cls, name, bases, attrs):
        __fields = dict({(field_name, attrs.pop(field_name)) for field_name, object in list(attrs.items()) if
                         isinstance(object, BaseField)})
        attrs['__fields'] = __fields
        new_class = type.__new__(cls, name, bases, attrs)
        return new_class


class Item(metaclass=ItemMeta):
    """
    Item class for each item
    """

    def __init__(self):
        self.ignore_item = False
        self.results = {}

    @classmethod
    async def _get_html(cls, html: str = '', url: str = '', **kwargs):
        if html or url:
            if url:
                sem = kwargs.pop('sem', None)
                request = Request(url, **kwargs)
                if sem:
                    _, response = await request.fetch_callback(sem=sem)
                else:
                    response = await request.fetch()
                html = response.html
            return etree.HTML(html)
        else:
            ValueError("html(url or html_etree) is expected")

    @classmethod
    async def _parse_html(cls, *, html_etree: etree._Element):
        if html_etree is None:
            raise ValueError("html_etree is expected")
        item_ins = cls()
        for field_name, field_value in getattr(item_ins, '__fields', {}).items():
            if not field_name.startswith('target_'):
                clean_method = getattr(item_ins, f'clean_{field_name}', None)
                value = field_value.extract(html_etree=html_etree)
                if clean_method is not None:
                    try:
                        value = await clean_method(value)
                    except TypeError:
                        raise InvalidFuncType('<Item: clean_method must be a coroutine function>')
                    except IgnoreThisItem:
                        item_ins.ignore_item = True

                setattr(item_ins, field_name, value)
                item_ins.results[field_name] = value
        return item_ins

    @classmethod
    async def get_item(cls, *,
                       html: str = '',
                       url: str = '',
                       html_etree: etree._Element = None,
                       **kwargs) -> Any:
        if html_etree is None:
            html_etree = await cls._get_html(html, url, **kwargs)

        return await cls._parse_html(html_etree=html_etree)

    @classmethod
    async def get_items(cls, *,
                        html: str = '',
                        url: str = '',
                        html_etree: etree._Element = None,
                        **kwargs):
        if html_etree is None:
            html_etree = await cls._get_html(html, url, **kwargs)
        items_field = getattr(cls, '__fields', {}).get('target_item', None)
        if items_field:
            items_field.many = True
            items_html_etree = items_field.extract(html_etree=html_etree, is_source=True)
            if items_html_etree:
                for each_html_etree in items_html_etree:
                    item = await cls._parse_html(html_etree=each_html_etree)
                    if not item.ignore_item:
                        yield item
            else:
                raise ValueError("Get target_item's value error!")
        else:
            raise ValueError("target_item is expected")

    def __str__(self):
        return f"<Item {self.results}>"



#!/usr/bin/env python
"""
 Created by howie.hu at 2019/1/28.
"""
import asyncio

from lxml import etree
from aiohttp.cookiejar import SimpleCookie

from ruia import Request


async def hello(response):
    return 'hello ruia'


sem = asyncio.Semaphore(3)
params = {
    "name": "ruia"
}
request = Request('http://www.httpbin.org/get',
                  method='GET',
                  metadata={'hello': 'ruia'},
                  params=params,
                  callback=hello)
_, response = asyncio.get_event_loop().run_until_complete(request.fetch_callback(sem))


def test_response():
    url = response.url
    method = response.method
    encoding = response.encoding
    html = response.html
    metadata = response.metadata
    cookies = response.cookies
    history = response.history
    headers = response.headers
    status = response.status
    html_etree = response.html_etree

    text = asyncio.get_event_loop().run_until_complete(response.text())
    json = asyncio.get_event_loop().run_until_complete(response.json())
    read = asyncio.get_event_loop().run_until_complete(response.read())

    assert url == 'http://www.httpbin.org/get'
    assert method == 'GET'
    assert encoding == 'utf-8'
    assert metadata == {'hello': 'ruia'}
    assert isinstance(html, str)
    assert isinstance(cookies, SimpleCookie)
    assert history == ()
    assert headers['Content-Type'] == 'application/json'
    assert status == 200
    assert response.ok == True
    assert isinstance(html_etree, etree._Element)
    assert isinstance(text, str)
    assert isinstance(json, dict)
    assert isinstance(read, bytes)

    assert str(response) == '<Response url[GET]: http://www.httpbin.org/get status:200>'


def test_callback():
    assert response.callback_result == 'hello ruia'
    response.callback_result = 'ruia'
    assert response.callback_result == 'ruia'


def test_index():
    assert response.index is None
    response.index = 'ruia'
    assert response.index == 'ruia'



#!/usr/bin/env python

import os

import lxml

from ruia import AttrField, TextField, HtmlField, RegexField
from ruia.field import NothingMatchedError

html_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'for_field_testing.html')
with open(html_path, mode='r', encoding='utf-8') as file:
    HTML = file.read()

html_etree = lxml.etree.HTML(HTML)


def test_css_select():
    field = TextField(css_select="head title")
    value = field.extract(html_etree=html_etree)
    assert value == "ruia"


def test_xpath_select():
    field = TextField(xpath_select='/html/head/title')
    value = field.extract(html_etree=html_etree)
    assert value == "ruia"


def test_attr_field():
    attr_field = AttrField(css_select="div.brand a", attr='href')
    value = attr_field.extract(html_etree=html_etree)
    assert value == "https://github.com"


def test_text_field_many():
    field = TextField(css_select="a.test_link", many=True)
    values = field.extract(html_etree=html_etree)
    assert values[2] == 'hello3 github.'


def test_attr_field_many():
    field = AttrField(css_select="a.test_link", attr="href", many=True)
    values = field.extract(html_etree=html_etree)
    assert len(values) == 5
    assert values[3] == "https://github.com/howie6879/ruia"


def test_text_field_not_exist():
    field = TextField(css_select="nothing matched")
    try:
        field.extract(html_etree=html_etree)
    except Exception as e:
        assert isinstance(e, NothingMatchedError)


def test_attr_field_not_exist():
    field = TextField(css_select="nothing matched")
    try:
        field.extract(html_etree=html_etree)
    except Exception as e:
        assert isinstance(e, NothingMatchedError)


def test_text_field_many_even_there_is_only_one_in_html():
    field = TextField(css_select="div.brand a", many=True)
    value = field.extract(html_etree=html_etree)
    assert value[0] == 'Github'


def test_attr_field_many_even_there_is_only_one_in_html():
    field = AttrField(css_select="div.brand a", attr="href", many=True)
    value = field.extract(html_etree=html_etree)
    assert value[0] == 'https://github.com'
    assert len(value) == 1


def test_text_field_with_default():
    field = TextField(css_select="div.brand b", default='nothing')
    value = field.extract(html_etree=html_etree)
    assert value == 'nothing'


def test_attr_field_with_default():
    field = AttrField(css_select="div.brand b", attr='href', default='nothing')
    value = field.extract(html_etree=html_etree)
    assert value == 'nothing'


def test_text_field_with_default_and_many():
    field = TextField(css_select="div.brand b", default="nothing", many=True)
    values = field.extract(html_etree=html_etree)
    assert values == ['nothing']


def test_attr_field_with_default_and_many():
    field = AttrField(css_select="div.brand b", attr="href", default="nothing", many=True)
    values = field.extract(html_etree=html_etree)
    assert values == ['nothing']


def test_html_field():
    field = HtmlField(css_select="div.brand a")
    assert field.extract(html_etree=html_etree) == '<a href="https://github.com">Github</a>'


def test_html_field_with_many():
    field = HtmlField(css_select="a.test_link", many=True)
    values = field.extract(html_etree=html_etree)
    assert len(values) == 5
    assert values[0] == '<a class="test_link" href="https://github.com/howie6879/">hello1 github.</a>\n'
    assert values[4] == '<a class="test_link" href="https://github.com/howie6879/">hello5 github.</a>\n' \
                        '    Some text outside.\n'


def test_re_field_with_one_group():
    field = RegexField(re_select='<title>(.*?)</title>')
    href = field.extract(html=HTML)
    assert href == 'ruia'


def test_re_field_with_no_group():
    field = RegexField(re_select='<title>.*?</title>')
    href = field.extract(html=HTML)
    assert href == '<title>ruia</title>'


def test_re_field_with_many_groups():
    field = RegexField(re_select='<h1><a href="(.*?)">(.*?)</a></h1>')
    href, text = field.extract(html=HTML)
    assert href == 'https://github.com'
    assert text == 'Github'


def test_re_field_with_named_groups():
    field = RegexField(re_select='<h1><a href="(?P<href>.*?)">(?P<text>.*?)</a></h1>')
    result = field.extract(html=HTML)
    assert result['href'] == 'https://github.com'
    assert result['text'] == 'Github'


def test_re_field_with_default():
    field = RegexField(re_select='nothing to match.', default='default value')
    result = field.extract(html=HTML)
    assert result == 'default value'


def test_re_field_get_nothing_with_no_default():
    field = RegexField(re_select='nothing to match.')
    try:
        field.extract(html=HTML)
    except Exception as e:
        assert isinstance(e, NothingMatchedError)


def test_re_field_with_many():
    field = RegexField(re_select='<a class="test_link" href="(.*?)">(.*?)</a>', many=True)
    matches = field.extract(html=HTML)
    assert len(matches) == 5
    href0, text0 = matches[0]
    href4, text4 = matches[4]
    assert href0 == 'https://github.com/howie6879/'
    assert text0 == 'hello1 github.'
    assert href4 == 'https://github.com/howie6879/'
    assert text4 == 'hello5 github.'


def test_re_field_in_dict_format_with_many():
    field = RegexField(re_select='<a class="test_link" href="(?P<href>.*?)">(?P<text>.*?)</a>', many=True)
    matches = field.extract(html=HTML)
    assert len(matches) == 5
    assert matches[0]['href'] == 'https://github.com/howie6879/'
    assert matches[0]['text'] == 'hello1 github.'
    assert matches[4]['href'] == 'https://github.com/howie6879/'
    assert matches[4]['text'] == 'hello5 github.'


def test_re_field_with_html_element():
    field = RegexField(re_select='<h1><a href="(?P<href>.*?)">(?P<text>.*?)</a></h1>')
    result = field.extract(html=html_etree)
    assert result['href'] == 'https://github.com'
    assert result['text'] == 'Github'



#!/usr/bin/env python

import json

from typing import Any, Callable, Optional

from lxml import etree

DEFAULT_JSON_DECODER = json.loads
JSONDecoder = Callable[[str], Any]


class Response(object):
    """
    Return a friendly response
    """

    def __init__(self, url: str, method: str, *,
                 encoding: str = '',
                 html: str = '',
                 metadata: dict,
                 cookies,
                 history,
                 headers: dict = None,
                 status: int = -1,
                 aws_json: Callable = None,
                 aws_read: Callable = None,
                 aws_text: Callable = None):
        self._callback_result = None
        self._encoding = encoding
        self._url = url
        self._method = method
        self._metadata = metadata
        self._html = html
        self._index = None
        self._cookies = cookies
        self._history = history
        self._headers = headers
        self._status = status

        self._aws_json = aws_json
        self._aws_read = aws_read
        self._aws_text = aws_text

    @property
    def callback_result(self):
        return self._callback_result

    @callback_result.setter
    def callback_result(self, value):
        self._callback_result = value

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    @property
    def encoding(self):
        return self._encoding

    @property
    def ok(self) -> bool:
        return self._status == 0 or 200 <= self._status <= 299

    @property
    def url(self):
        return self._url

    @property
    def method(self):
        return self._method

    @property
    def metadata(self):
        return self._metadata

    @property
    def html(self):
        return self._html

    @property
    def cookies(self):
        return self._cookies

    @property
    def history(self):
        return self._history

    @property
    def headers(self):
        return self._headers

    @property
    def status(self):
        return self._status

    @property
    def html_etree(self):
        html_etree = None
        if self.html:
            html_etree = etree.HTML(self.html)
        return html_etree

    async def json(self, *, encoding: str = None,
                   loads: JSONDecoder = DEFAULT_JSON_DECODER,
                   content_type: Optional[str] = 'application/json') -> Any:
        """Read and decodes JSON response."""
        return await self._aws_json(encoding=encoding,
                                    loads=loads,
                                    content_type=content_type)

    async def read(self) -> bytes:
        """Read response payload."""
        return await self._aws_read()

    async def text(self, *, encoding: Optional[str] = None,
                   errors: str = 'strict') -> str:
        """Read response payload and decode."""
        return await self._aws_text(encoding=encoding,
                                    errors=errors)

    def __str__(self):
        return f'<Response url[{self._method}]: {self._url} status:{self._status}>'
